@@ -11,6 +11,7 @@ from typing import Any, Callable, Dict, List, Optional, Tuple, Union, Set
 
 from .prefetch_generator import PrefetchGenerator
 from ..utils import is_available, width
+from ..utils._proxy_class import _proxy_class
 
 import numpy as np
 
@@ -22,7 +23,9 @@ class Dataset:
     """
     """
     
-    def __init__(self, data: List[Dict[str, Any]] = None, shuffle: bool = False, *args, **kwargs) -> None:
+    def __init__(self, data: List[Dict[str, Any]] = None, *args, **kwargs) -> None:
+        self._has_changed = False
+        self.data = _proxy_class(list, self)()
         if data is None:
             pass
         else:
@@ -35,71 +38,58 @@ class Dataset:
             else:
                 raise(ValueError("data는 list 타입이어야 합니다."))
 
-        self._data = data if data is not None else []
-        self._shuffle = shuffle
+        self.data.extend(data) if data is not None else None
+        self._num_data = None
+        self._counts = None
+    
+    def update_states(func):
+        def wrapper(self, *args, **kwargs):
+            if self._num_data is None or self._counts is None or self._has_changed:
+                self._get_counts()
+                self._num_data = len(self.data)
+                self._has_changed = False
+            return func(self, *args, **kwargs)
+        return wrapper            
+
+    def shuffle(self):
+        random.shuffle(self.data)
 
     def append(self, data: Dict[str, Any]) -> None:
-        self._data.append(data)
-
-    add = append
+        self.data.append(data)
 
     def drop(self, condition: function) -> Dataset:
         return self.filter(lambda x: not condition(x))
 
     def filter(self, condition: function) -> Dataset:
-        return Dataset(list(filter(condition, self._data)))
+        return Dataset(list(filter(condition, self.data)))
 
     def merge(self, key: str, values: Union[List[str], Tuple[str]], sort: bool = True) -> Dataset:
         if isinstance(values, (list, tuple)):
             if sort:
                 values = sorted([str(v) for v in values])
             new_values = "+".join(values)
-            return Dataset([{k: (new_values if (key == k and d[k] in values) else d[k]) for k in d.keys()} for d in self._data])
+            return Dataset([{k: (new_values if (key == k and d[k] in values) else d[k]) for k in d.keys()} for d in self.data])
         else:
             raise TypeError("unsupported operand type(values) for merge")
 
-    def __add__(self, other: Dataset) -> Dataset:
-        cls = Dataset(shuffle=self._shuffle)
-        cls._data.extend(copy.deepcopy(self._data))
-        cls._data.extend(copy.deepcopy(other._data))
-        return cls
-
-    def __radd__(self, other: Dataset) -> Dataset:
-        return self + other
-
-    def __getitem__(self, index: Union[int, slice]) -> Dataset:
-        cls = Dataset()
-        data = self._data[index]
-        if isinstance(data, dict):
-            cls._data.append(data)
-        else:
-            cls._data.extend(data)
-        return cls
-
-    def __len__(self) -> int:
-        return self.num_data
-
-    @property
-    def data(self) -> List[Dict[str, Any]]:
-        if self._shuffle:
-            random.shuffle(self._data)
-        return self._data
-
-    @property
-    def num_data(self):
-        return len(self._data)
+    def _get_counts(self):
+        _counts = defaultdict(lambda: defaultdict(int))
+        for d in self.data:
+            for k, v in d.items():
+                _counts[k][v] += 1
+        self._counts = {k1: dict(v1) for k1, v1 in _counts.items()}
 
+    @update_states
     def counts(self, keys: Optional[Union[List[str], Tuple[str], Set[str], str]] = None):
         if keys is not None and isinstance(keys, str):
             keys = set([keys])
-        _counts = defaultdict(lambda: defaultdict(int))
-        for d in self._data:
-            for k, v in d.items():
-                if keys is not None and k in keys:
-                    _counts[k][v] += 1
-                else:
-                    _counts[k][v] += 1
-        return {k1: dict(v1) for k1, v1 in _counts.items()}
+        if keys is not None:
+            tmp = {}
+            for k in keys:
+                tmp[k] = self._counts[k]
+            return tmp
+        else:
+            return self._counts
 
     def combination_counts(self, keys: Union[List[str], Tuple[str], Set[str]], views: Optional[Union[List[str], Tuple[str], Set[str], str]] = None) -> List[Dict[str, Any]]:
         def _cond(keys, values):
@@ -140,18 +130,6 @@ class Dataset:
                         _tmp[v] = list(set(_ds_counts[v]))
                 _counts.append(_tmp)
                 _tmp["count"] = ds.num_data
-            
-        # max_lengths = [0 for _ in range(len(keys))]
-        # for c in combinations:
-        #     for i in range(len(keys)):
-        #         if max_lengths[i] < width(c[i]):
-        #             max_lengths[i] = width(c[i])
-        # _counts = {}
-        # for c in combinations:
-        #     name = ", ".join([f"{k.title()}: {v:{max_lengths[i]}s}" for i, (k, v) in enumerate(zip(keys, c))])
-        #     n = self.filter(_cond(keys, c)).num_data
-        #     if n > 0:
-        #         _counts[name] = n
         return _counts
         
     def print_counts(self, max_display: int = 10, keys: Optional[Union[List[str], Tuple[str], Set[str], str]] = None, print_all: bool = False):
@@ -175,11 +153,39 @@ class Dataset:
         _repr = re.sub(r"^", " " * 4, _repr, 0, re.M)
         print(f"Dataset: {self.num_data}\n{_repr}")
 
+    def __add__(self, other: Dataset) -> Dataset:
+        cls = Dataset(shuffle=self._shuffle)
+        cls.data.extend(copy.deepcopy(self.data))
+        cls.data.extend(copy.deepcopy(other.data))
+        return cls
+
+    def __radd__(self, other: Dataset) -> Dataset:
+        if not isinstance(other, Dataset):
+            other = Dataset()
+        return self + other
+
+    def __getitem__(self, index: Union[int, slice]) -> Dataset:
+        cls = Dataset()
+        data = self.data[index]
+        if isinstance(data, dict):
+            cls.data.append(data)
+        else:
+            cls.data.extend(data)
+        return cls
+
+    def __len__(self) -> int:
+        return self.num_data
+
+    @property
+    @update_states
+    def num_data(self):
+        return self._num_data
+
     def __repr__(self):
         return f"Dataset: {self.num_data}\n"
 
     def __iter__(self):
-        return iter(self._data)
+        return iter(self.data)
 
 
 class DataLoader:
@@ -198,7 +204,7 @@ class DataLoader:
             if is_available("torch"):
                 import torch
         self._padding_value = padding_value
-        self._gen = PrefetchGenerator(self.dataset._data, num_prefetch=self._prefetch_factor*self._batch_size, num_workers=self._num_workers, shuffle=self._shuffle, processing_func=self._collate_fn, name="DataLoader")
+        self._gen = PrefetchGenerator(self.dataset.data, num_prefetch=self._prefetch_factor*self._batch_size, num_workers=self._num_workers, shuffle=self._shuffle, processing_func=self._collate_fn, name="DataLoader")
 
     def as_format(self, batch):
         if is_available("numpy"):
